# https://medium.com/bootdotdev/aes-256-cipher-python-cryptography-examples-b877b9d2e45e
# AES 256 encryption/decryption using pycrypto library
import base64
import hashlib
import os
import sys

from Crypto import Random
from Crypto.Cipher import AES


class PyAES256:
    # constructor
    def __init__(self):
        pass
        
    # pad with spaces at the end of the text
    # becouse AES needs 16 byte blocks
    def pad(self, s):
        block_size = 16
        remainder = len(s) % block_size
        padding_needed = block_size - remainder
        return s + padding_needed * ' '

    # remove the extra spaces at the end
    def unpad(self, s): 
        return s.rstrip()

    def bytes_to_str(self, b):    
        return ''.join([chr(c) for c in b])

    def str_to_bytes(self, b):
        return bytes.fromhex(''.join([hex(ord(c)).replace('x','0')[-2:] for c in b]))    

    # do encryption  
    # @staticmethod  
    def encrypt(self, plain_text, password):
        # generate a random salt    
        salt = os.urandom(AES.block_size)
        # print('salt: ', salt)

        # generate a random iv
        iv = Random.new().read(AES.block_size)
        # print('increment vector: ', iv)

        # use the Scrypt KDF to get a private key from the password
        private_key = hashlib.scrypt(password.encode(), salt=salt, n=2**14, r=8, p=1, dklen=32)
        # print('private key: ', private_key)

        # pad text with spaces to be valid for AES CBC mode
        padded_text = self.pad(plain_text)
                
        # create cipher config
        cipher_config = AES.new(private_key, AES.MODE_CBC, iv)    

        ciper = cipher_config.encrypt(bytes(padded_text,'utf-8'))
        # print('Chipper: ', ciper)
        
        # return a dictionary with the encrypted text
        ciper_text = base64.b64encode(ciper)
        # print('Chipper Text: ', ciper_text)

        url_text = self.bytes_to_str(base64.urlsafe_b64encode(ciper_text))

        return {
            # 'cipher_text': ciper_text,          # convert to url safe
            'url': url_text,                    # set as url encryption
            'salt': base64.b64encode(salt),     # salt & iv save to session 
            'iv': base64.b64encode(iv),                     
        }    

    # do decription
    # @staticmethod
    def decrypt(self, url, salt, iv, password):
        # decode the dictionary entries from base64
        salt = base64.b64decode(salt)
        # print('salt: ', salt)

        ciper_text = base64.b64decode(base64.urlsafe_b64decode(self.str_to_bytes(url)))
        # print('chipper text: ', ciper_text)

        iv = base64.b64decode(iv)
        # print('increment vector: ', iv)

        # url_text = enc_dict['url']
        # print('URL: ', url_text)

        # generate the private key from the password and salt
        private_key = hashlib.scrypt(password.encode(), salt=salt, n=2**14, r=8, p=1, dklen=32)
        # print('private key: ', private_key)

        # create the cipher config
        cipher = AES.new(private_key, AES.MODE_CBC, iv)
        # print('chipper: ', cipher)

        # decrypt the cipher text
        decrypted = cipher.decrypt(ciper_text)
        # print('decript: ', decrypted)

        # unpad the text to remove the added spaces
        original = self.unpad(decrypted)
        # print('unpad: ', original)

        return original


# test module level
if __name__=='__main__':
    print('Begin Test ENCRYPT')
    print('------------------')
    secret_text = input("Secret Text: ")
    password = 'g_7me8rl2m#a_h2oresgt2#ni=3_4*!ai*=rtsq)yi!g7_5-51'     

    lib = PyAES256()
    res_enc = lib.encrypt(secret_text, password)
    print('Encrypt result:', res_enc)

    print('')
    print('Begin Test DECRYPT')
    print('------------------')
    
    url = res_enc['url']
    salt = res_enc['salt']
    iv = res_enc['iv']
    print('Parameter')
    print('PASSWORD:', password)
    print('URL:', url)
    print('SALT:', salt)
    print('IV:', iv)    

    res_dec = lib.decrypt(url, salt, iv, password)
    print('Decrypt result:', bytes.decode(res_dec))
    print('')
    



import math

class PyAvalanche:
    '''
        Create avalance effect for cryptograph testing
        ref : https://www.researchgate.net/post/What_is_the_avalanche_effect_in_cryptography_How_can_we_measure_it
              https://www.educative.io/answers/how-to-convert-string-to-binary-in-python
    '''
    # constructor
    def __init__(self):
        pass

    def str_to_binary(self, a):
        l,m=[],[]
        for i in a:
            l.append(ord(i))

        for i in l:
            m.append(int(bin(i)[2:]))

        return m

    def binary_to_str(self, a):
        l=[]
        m=""
        for i in a:
            b=0
            c=0
            k=int(math.log10(i))+1
            for j in range(k):
                b=((i%10)*(2**j))   
                i=i//10
                c=c+b
            l.append(c)

        for x in l:
            m=m+chr(x)

        return m

    def avalanche_effect(self, chipper1, chipper2):
        '''
            chipper1 and chipper2 in string
        '''
        a = self.str_to_binary(chipper1)
        b = self.str_to_binary(chipper2)

        a = int(''.join([str(i) for i in a]))
        b = int(''.join([str(i) for i in b]))

        # print ('chipper a =',a)
        # print ('chipper b =',b)

        # print bitwise XOR operation
        a_xor_b = a^b
        # print("a ^ b =", a_xor_b)

        # Base 2(binary)
        bin_a_xor_b = bin(a_xor_b)
        # print("the result in decimal =", bin_a_xor_b)

        # next steps to count 1s in binary number
        one_count = 0
        for i in bin_a_xor_b:
            if i == "1":
                one_count+=1
        # print ("the 1s numbers is=", one_count)

        # next steps to calculate equation of avalanche effect
        len_a = len(bin(a))
        len_b = len(bin(b))
        # print(len_a, len_b)

        # if to ensure divide by the longest binary string
        if (len_a) >= (len_b):
            AVA = (one_count/ len_a) * 100
        else:
            AVA = (one_count/ len_b) * 100

        # print ("avalanche effect =", AVA, "%")

        return AVA


# test module level
if __name__=='__main__':
    '''
        Chipper1 is result from any encryption, first run
        Chipper2 is result from same encryption, second run
    '''
    chipper1 = 'Nk9iL05BQmM5bnRES3drVGc0NHZRdz09'
    chipper2 = 'Z3VFbi9lWDJtS1B5UzE4TGVsMzRHQT09'
    print('Chipper 1 = ', chipper1)
    print('Chipper 2 = ', chipper2)
    ava = PyAvalanche()
    result = ava.avalanche_effect(chipper1, chipper2)

    print(result,'%')
    print("If more than 50% output bits have changed then the algo have good avalanche effect.")